// owl_comm.h

#ifndef OWL_COMM_H
#define OWL_COMM_H

#include <iostream>
#include <map>

namespace OWL {

  //// CommHeader ////

  struct CommHeader {
    char magic[8];
    int frame;
    unsigned char data[];
  };

  //// CommPacket ////

  struct CommPacket {
    enum { Hub = 0x1, RX = 0x2 };
    unsigned short type;
    unsigned short size;
    unsigned char data[];
  };

  //// HubPacket ////

  struct HubPacket {
    unsigned char id[8];
    unsigned char frameCount;
    unsigned char frameFlags;
    unsigned char pad[2];
    unsigned int frameCountTTL;
  };

  //// RXPacket ////

  struct RXPacket {

    enum { State = 0x1, RFScan = 0x2, RFData = 0x3 };

    enum {
      CAMFLAGS_STATE = 1<<7,
      CAMFLAGS_RXSLOT = 1<<6,
      CAMFLAGS_RF_PIPE = 3<<0,
    };

    enum {
      PACKET_TYPE_MASK = 1<<7,
      PACKET_TYPE_SCAN = 0<<7,
      PACKET_TYPE_DATA = 1<<7
    };

    unsigned char camid[4];
    unsigned char hwid[2];
    unsigned char id;
    unsigned char status;
    unsigned char rxMeta[4];
    unsigned char camFlags;
    unsigned char camStatus;
    unsigned char data[18];

    operator int() const { return ((hwid[0] << 8) | (hwid[1] << 0)) >> 3; }

    int type() const;
  };

  //// CamStatePacket ////

  struct CamStatePacket {
    unsigned char ledSlot; // backup
    unsigned char rxSlot[3]; // 0:config, 1:state, 2:scan addr
    unsigned char txFlags[2]; // handshake for camera tx's
    unsigned char data[12];
  };

  //// RFScanPacket ////

  struct RFScanPacket {
    unsigned char type;
    unsigned char hwid[2];
    unsigned char id;
    unsigned char rxAddr[2];
    unsigned char signalStatus;
    unsigned char batteryStatus;
    unsigned char encodeStatus;
    unsigned char readoutAddr;
    unsigned char readout[8];
  };

  //// RFDataPacket ////

  struct RFDataPacket {
    unsigned char type;
    unsigned char retry;
    unsigned char data[16];
  };

  //// CommParser ////

  class CommParser {
  protected:

    const CommHeader *header;
    const unsigned char *head, *end;

  public:

    CommParser(const char *data, size_t count);

    operator bool() const { return header; }

    int frame() const;

    const CommPacket* next();
  };

  //// RFDevice ////

  struct RFDevice {

    unsigned short hwid;
    unsigned char id;
    unsigned char status;

    int lastUpdated, lastChanged;

    RFDevice();
    RFDevice(const RXPacket &rx, int frame);

    void update(const RXPacket &rx, int frame);

    bool isValid() const { return lastChanged > -1; }
  };

  //// RFDevices ////

  class RFDevices : public std::map<int,RFDevice> {
  public:

    typedef std::map<int,RFDevice>::iterator iterator;
    typedef std::map<int,RFDevice>::const_iterator const_iterator;

    int verbose;

    RFDevices();

    void update(const RXPacket &rx, int frame, int timeout);
    void info(int frame=-1) const;
  };

  ////

  //// RXPacket ////

  inline int RXPacket::type() const
  {
    if(camFlags & CAMFLAGS_STATE) return State;
    if((camFlags & CAMFLAGS_RF_PIPE) != 0) return 0;
    return (data[0] & PACKET_TYPE_MASK) == PACKET_TYPE_SCAN ? RFScan : RFData;
  }

  //// CommParser ////

  inline CommParser::CommParser(const char *data, size_t count) : header(0), head(0), end(0)
  {
    if(count < sizeof(CommHeader) || memcmp(data, "COMM2.00", 8) != 0) return;
    header = (const CommHeader*)data;
    head = header->data;
    end = (const unsigned char*)data + count;
  }

  inline int CommParser::frame() const { return header ? header->frame : -1; }

  inline const CommPacket* CommParser::next()
  {
    if(!header || head + sizeof(CommPacket) > end) return 0;
    const CommPacket *p = (const CommPacket*)head;
    head += sizeof(CommPacket) + p->size;
    return head <= end ? p : 0;
  }

  //// RFDevice ////

  inline RFDevice::RFDevice() : hwid(0), id(0), status(0), lastUpdated(-1), lastChanged(-1)
  {
  }

  inline RFDevice::RFDevice(const RXPacket &rx, int frame) :
    hwid(rx>>3), id(rx.id), status(rx.status), lastUpdated(frame), lastChanged(frame)
  {
  }
  
  inline void RFDevice::update(const RXPacket &rx, int frame)
  {
    if(id != rx.id || status != rx.status) lastChanged = frame;
    id = rx.id;
    status = rx.status;
    lastUpdated = frame;
  }

  //// RFDevices ////

  inline RFDevices::RFDevices() : verbose(1)
  {
  }

  inline void RFDevices::update(const RXPacket &rx, int frame, int timeout)
  {
    if((rx.type() == RXPacket::RFScan || rx.type() == RXPacket::RFData) && rx != 0)
      {
        // find or create device to update
        iterator i = find(rx);
        if(i == end())
          {
            i = insert(std::make_pair<>(rx, RFDevice(rx, frame))).first;
            if(verbose) std::cout << "new device: " << std::hex << i->second.hwid << std::dec << std::endl;
          }
        else i->second.update(rx, frame);
      }

    // invalidate stale devices
    for(iterator i = begin(); timeout > 0 && i != end(); i++)
      if(i->second.isValid() && i->second.lastUpdated + timeout < frame)
        {
          if(verbose) std::cout << "remove device: " << std::hex << i->second.hwid << std::dec << std::endl;
          // invalidate
          i->second.status = 0;
          i->second.lastUpdated = -1;
          i->second.lastChanged = -1;
        }
  }

  inline void RFDevices::info(int frame) const
  {
    if(empty()) return;
    if(frame < 0)
      {
        size_t count = 0;
        for(const_iterator i = begin(); i != end(); i++)
          if(i->second.isValid()) count++;
        std::cout << count << " device(s):" << std::endl;
      }
    for(const_iterator i = begin(); i != end(); i++)
      {
        const RFDevice &d = i->second;
        if(d.isValid())
          {
            if(frame > -1 && frame != d.lastChanged) continue;
            std::cout << "device hwid=0x" << std::hex << d.hwid
                      << " id=" << std::dec << (int)d.id
                      << " status=0x" << std::hex << (int)d.status
                      << std::dec << std::endl;
          }
      }
  }

  ////

} // namespace OWL

#endif // OWL_COMM_H
