// owl_file.h
// owl file utilities

#ifndef OWL_FILE_H
#define OWL_FILE_H

#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <assert.h>
#include <errno.h>

#include "owl.h"
#include "owl_peaks.h"
#include "owl_planes.h"

using namespace std;

inline ostream &operator<<(ostream &out, const OWLMarker &m)
{
  out << "0x" << setfill('0') << hex << setw(4) << m.id << dec;
  out << ", " << setfill(' ') << setw(14) << m.cond << ", ";
  out << setw(14) << m.x << " " 
      << setw(14) << m.y << " "
      << setw(14) << m.z;
  return out;
}

inline ostream &operator<<(ostream &out, const OWLRigid &r)
{
  out << "0x" << setfill('0') << hex << setw(4) << r.id << dec;
  out << ", " << setfill(' ') << setw(14) << r.cond << ",";
  for(int i = 0; i < 7; i++) out << " " << setw(14) << r.pose[i];
  return out;
}

inline ostream &operator<<(ostream &out, const OWLCamera &c)
{
  out << "0x" << setfill('0') << hex << setw(4) << c.id << dec;
  out << ", " << setfill(' ') << setw(14) << c.cond << ",";
  for(int i = 0; i < 7; i++) out << " " << setw(14) << c.pose[i];
  return out;
}

inline ostream &operator<<(ostream &out, const OWLDetectors &d)
{
  out << "0x" << setfill('0') << hex << setw(4) << d.id;
  out << "," << setfill(' ');
  for(size_t i = 0; i < 8; i++) out << " " << setw(14) << d.detectors[i];
  out << dec;
  return out;
}

inline int save(const char *filename, const vector< vector<OWLMarker> > &recorded, 
		const vector<OWLCamera> *cameras, int frame_offset=0, float freq=OWL_MAX_FREQUENCY)
{
  assert(frame_offset >= 0);
  ofstream out(filename);
  
  if(!out.is_open())
    {
      cerr << "open(" << filename << "): " << strerror(errno) << endl;
      return 0;
    }

  out.setf(ios::fixed, ios::floatfield);

  out << "# " << recorded.size()-frame_offset << " frames" << endl;

  out.precision(2);
  out << "q " << freq << endl;

  out.precision(6);
  // dump cameras
  if(cameras)
    {
      for(size_t i = 0; i < cameras->size(); i++)
	out << "c " << (*cameras)[i] << endl;
      out << endl;
    }

  // dump markers
  for(size_t i = frame_offset; i < recorded.size(); i++)
    {
      const vector<OWLMarker> &m = recorded[i];

      assert(m.size());
      
      out << "f " << m[0].frame-frame_offset << endl;

      for(size_t j = 0; j < m.size(); j++)
        out << "m " << m[j] << endl;
      out << endl;
    }

  cout << "wrote to " << filename << endl;
  return 1;
}

inline int load(const char *filename, vector< vector<OWLMarker> > &recorded, 
		vector<OWLCamera> *cameras, float &freq)
{
  ifstream in(filename);

  if(!in.is_open())
    {
      cerr << "open(" << filename << "): " << strerror(errno) << endl;
      return 0;
    }

  int lines = 0;
  char str[1024];
  int frame;
  OWLMarker m;
  OWLCamera c;
  //OWLRigid r;
  vector<OWLMarker> mks;
  while(1)
    {
      in.getline(str, 1024);

      if(in.eof()) break;

      lines++;
      if(in.gcount() <= 1) continue;

      switch(str[0])
	{
	case '#': // comment
	  // ignore
	  break;
        case 'q': // freq
          cout << "eval: " << str << endl;
          if(sscanf(str, "q %f", &freq) != 1) goto error;
          break;
	case 'f': // frame
	  if(sscanf(str, "f %d", &frame) != 1) goto error;
	  if(mks.size()) recorded.push_back(mks);
	  mks.clear();
	  break;
	case 'c': // camera
	  if(sscanf(str, "c %x, %f, %f %f %f %f %f %f %f", &c.id, &c.cond, 
		    &c.pose[0], &c.pose[1], &c.pose[2], &c.pose[3], 
		    &c.pose[4], &c.pose[5], &c.pose[6]) != 9) goto error;
	  c.flag = 0;
	  if(cameras) cameras->push_back(c);
	  break;
	case 'm': // marker
	  if(sscanf(str, "m %x, %f, %f %f %f", &m.id, &m.cond, &m.x, &m.y, &m.z) != 5) goto error;
	  m.frame = frame;
	  m.flag = 0;
	  mks.push_back(m);
	  break;
	case 'r': // rigid body
	  break;
	case 'p': // planes
	  break;
	default:
	  cerr << "unknown token '" << str[0] << "' in file " << filename << endl;
	  return 0;
	error:
	  cerr << "error reading " << filename << ":" << lines << endl;
	  return 0; 
	}
    }
  
  if(mks.size()) recorded.push_back(mks);

  cout << "read " << recorded.size() << " frames from " << filename << endl;

  return 1;
}

inline int load(const char *filename, vector< vector<OWLMarker> > &recorded, 
		vector<OWLCamera> *cameras)
{
  ifstream in(filename);

  if(!in.is_open())
    {
      cerr << "open(" << filename << "): " << strerror(errno) << endl;
      return 0;
    }

  int lines = 0;
  char str[1024];
  int frame;
  OWLMarker m;
  OWLCamera c;
  //OWLRigid r;
  vector<OWLMarker> mks;
  while(1)
    {
      in.getline(str, 1024);

      if(in.eof()) break;

      lines++;
      if(in.gcount() <= 1) continue;

      switch(str[0])
	{
	case '#': // comment
	  // ignore
	  break;
	case 'f': // frame
	  if(sscanf(str, "f %d", &frame) != 1) goto error;
	  if(mks.size()) recorded.push_back(mks);
	  mks.clear();
	  break;
	case 'c': // camera
	  if(sscanf(str, "c %x, %f, %f %f %f %f %f %f %f", &c.id, &c.cond, 
		    &c.pose[0], &c.pose[1], &c.pose[2], &c.pose[3], 
		    &c.pose[4], &c.pose[5], &c.pose[6]) != 9) goto error;
	  c.flag = 0;
	  if(cameras) cameras->push_back(c);
	  break;
	case 'm': // marker
	  if(sscanf(str, "m %x, %f, %f %f %f", &m.id, &m.cond, &m.x, &m.y, &m.z) != 5) goto error;
	  m.frame = frame;
	  m.flag = 0;
	  mks.push_back(m);
	  break;
	case 'r': // rigid body
	  break;
	case 'p': // planes
	  break;
	default:
	  cerr << "unknown token '" << str[0] << "' in file " << filename << endl;
	  return 0;
	error:
	  cerr << "error reading " << filename << ":" << lines << endl;
	  return 0; 
	}
    }
  
  if(mks.size()) recorded.push_back(mks);

  cout << "read " << recorded.size() << " frames from " << filename << endl;

  return 1;
}

struct OWLFrame {
  
  int frame;

  vector<OWLMarker> markers;
  vector<OWLRigid> rigids;
  vector<OWLDetectors> detectors;
  vector<OWLPlane> planes;

  OWLFrame(int frame=-1) : frame(frame) { }

  operator void*()
  { 
    if(markers.size() || rigids.size()) return (void*)(-1);
    if(planes.size() || detectors.size()) return (void*)(-1);
    
    return (void*)0;
  }
  
  void clear()
  {
    markers.clear();
    rigids.clear();
    detectors.clear();
    planes.clear();
  }
};


inline int save(const char *filename, const vector<OWLFrame> &recorded, 
		const vector<OWLCamera> *cameras, int frame_offset=0)
{
  assert(frame_offset >= 0);
  ofstream out(filename);
  
  if(!out.is_open())
    {
      cerr << "open(" << filename << "): " << strerror(errno) << endl;
      return 0;
    }

  out.setf(ios::fixed, ios::floatfield);

  out << "# " << recorded.size()-frame_offset << " frames" << endl;

  // dump cameras
  if(cameras)
    {
      for(size_t i = 0; i < cameras->size(); i++)
	out << "c " << (*cameras)[i] << endl;
      out << endl;
    }

  // dump markers
  for(size_t i = frame_offset; i < recorded.size(); i++)
    {
      const vector<OWLMarker> &m = recorded[i].markers;
      const vector<OWLDetectors> &d = recorded[i].detectors;

      assert(m.size());
      
      out << "f " << m[0].frame-frame_offset << endl;
      
      for(size_t j = 0; j < m.size(); j++)
	out << "m " << m[j] << endl;
      for(size_t j = 0; j < d.size(); j++)
	out << "d " << d[j] << endl;
      out << endl;
    }

  cout << "wrote to " << filename << endl;
  return 1;
}

inline int load(const char *filename, vector<OWLFrame> &recorded, 
		vector<OWLCamera> *cameras)
{
  ifstream in(filename);

  if(!in.is_open())
    {
      cerr << "open(" << filename << "): " << strerror(errno) << endl;
      return 0;
    }

  int ret = 0;
  int lines = 0;
  char str[1024];
  //int frame;
  OWLMarker m;
  OWLCamera c;
  //OWLRigid r;
  //vector<OWLMarker> mks;
  OWLDetectors d;
  OWLFrame frame;
  while(1)
    {
      in.getline(str, 1024);

      if(in.eof()) break;

      lines++;
      if(in.gcount() <= 1) continue;

      switch(str[0])
	{
	case '#': // comment
	  // ignore
	  break;
	case 'f': // frame
	  if((ret = sscanf(str, "f %d", &frame.frame)) != 1) goto error;
 	  recorded.push_back(frame);
	  frame.clear();
	  break;
	case 'c': // camera
	  if(sscanf(str, "c %x, %f, %f %f %f %f %f %f %f", &c.id, &c.cond, 
		    &c.pose[0], &c.pose[1], &c.pose[2], &c.pose[3], 
		    &c.pose[4], &c.pose[5], &c.pose[6]) != 9) goto error;
	  c.flag = 0;
	  if(cameras) cameras->push_back(c);
	  break;
	case 'm': // marker
	  if(sscanf(str, "m %x, %f, %f %f %f", &m.id, &m.cond, &m.x, &m.y, &m.z) != 5) goto error;
	  m.frame = frame.frame;
	  m.flag = 0;
	  frame.markers.push_back(m);
	  break;
	case 'r': // rigid body
	  break;
	case 'p': // planes
	  break;
	case 'd': // detectors
	  if((ret = sscanf(str, "d %x, %x %x %x %x %x %x %x %x", &d.id, 
			   &d.detectors[0], &d.detectors[1], &d.detectors[2], &d.detectors[3],
			   &d.detectors[4], &d.detectors[5], &d.detectors[6], &d.detectors[7])) != 9)
	    goto error;
	  d.frame = frame.frame;
	  frame.detectors.push_back(d);
	  break;
	default:
	  cerr << "unknown token '" << str[0] << "' in file " << filename << endl;
	  return 0;
	error:
	  cerr << "error reading " << filename << ":" << lines << endl;
	  cerr << "[" << str << "] ret=" << ret << endl;
	  return 0; 
	}
    }
  
  if(frame) recorded.push_back(frame);

  cout << "read " << recorded.size() << " frames from " << filename << endl;

  return 1;
}

#endif // OWL_FILE_H
