#include "vrpn_Tracker_PhaseSpace.h"

#include <errno.h>
#include <iostream>
#include <algorithm>
#include <string>

#define MM_TO_METERS (0.001)

#ifdef VRPN_INCLUDE_PHASESPACE
//#define DEBUG
//#define BENCHMARK

#define MAXMARKERS 512
#define MAXRIGIDS 128
#define MAXCAMERAS 128
#define MAXDETECTORS 512
#define MAXPLANES 1024
#define MAXPEAKS 1024
#define MAXIMAGES 1024
#define MSGBUFSIZE 1024
#define COMMDATASIZE (64*1024)

#define MAX_STYLUS 256
#define SCALARS_PER_STYLUS 4
#define MAXSCALARS 1024

#define STYLUS_WIDTH 38.3
#define _ACCL (6.125E-4)
#define _GYRO (-5.326E-4)

//
vrpn_Tracker_PhaseSpace::StylusWrapper::StylusWrapper(const vrpn_Tracker_PhaseSpace::StylusWrapper &o)
{
  memcpy(this, &o, sizeof(o));
}

//
vrpn_Tracker_PhaseSpace::StylusWrapper::StylusWrapper(int hwid, int id, float width=STYLUS_WIDTH)
{
  this->hwid = hwid;
  this->id = id;
  std::fill(markers, markers+2, -1);
  std::fill(calib, calib+SC_WIDTH, 0);
  calib[0] = SC_WIDTH;
  calib[4] = calib[5] = calib[6] = _ACCL;
  calib[10] = calib[11] = calib[12] = _GYRO;
  calib[14] = width; // marker distance
}

//
bool vrpn_Tracker_PhaseSpace::StylusWrapper::operator==(int n) const
{
  return hwid == n;
}

// comm v1
struct BaseData {

  int32_t baseDataFlag;
  int32_t baseDataStamp;

  unsigned char magic_num[4];  // 0x60 0x06 0x07 0x70
  unsigned char pad0[8]; // reserved
  unsigned char flag;
  unsigned char frame[3]; // 24-bit, LSB
  unsigned char ttl; // ttl input
  unsigned char pad1; // reserved
  unsigned char data[46]; // ltc|vitc|irig data
};


//
int stylus_to_rigids(std::vector<vrpn_Tracker_PhaseSpace::StylusWrapper> &stylus, std::vector<OWLRigid> &rigids)
{
  if(stylus.size())
    {
      rigids.resize(MAX_STYLUS); // bug in libowlhybrid(?)
      owlHybridGetRigids(&rigids[0], rigids.size(), &stylus[0], stylus.size());
    }
  return stylus.size();
}


//
inline void frame_to_timeval(int64_t fr, double freq, struct timeval &tv)
{
  int64_t t = fr * 1e6 / freq;
  const int64_t d = 1000000;
  tv.tv_sec = t / d;
  tv.tv_usec = t % d;
  return;
}

//
bool process_basedata(const std::vector<unsigned char> &commdata, double freq, struct timeval &timestamp)
{
  if(commdata.size() < sizeof(BaseData) || freq <= 0) return false;

  const char magic[] = { 0x60, 0x06, 0x07, 0x70 };
  const BaseData *data = (const BaseData*)&commdata[0];

  if(memcmp(magic, data->magic_num, 4) != 0) return false;

  // check for data->flag?
  int64_t fc = ((data->frame[2] << 16) | (data->frame[1] << 8) | data->frame[0]) - 2;
  frame_to_timeval(fc, freq, timestamp);
  return true;
}


//
bool process_hubdata(const std::vector<unsigned char> commdata, double freq, struct timeval &timestamp)
{
  if(commdata.empty() || freq <= 0) return false;

  if(OWL::CommParser comm = OWL::CommParser((const char*) &*commdata.begin(), commdata.size()))
    while(const OWL::CommPacket *packet = comm.next())
      if(packet->type == OWL::CommPacket::Hub)
        {
          OWL::HubPacket *h = (OWL::HubPacket*)packet->data;
          if(h->frameCountTTL > 2)
            {
              int64_t fc = h->frameCountTTL - 2;
              frame_to_timeval(fc, freq, timestamp);
              return true;
            }
        }
  return false;
}

// read through commdata for hybrid tracking data
int vrpn_Tracker_PhaseSpace::update_hybrid(const std::vector<unsigned char> &data)
{
  // read through commdata and allocate new styluses
  if(OWL::CommParser comm = OWL::CommParser((const char*) &*data.begin(), data.size()))
    {
      while(const OWL::CommPacket *packet = comm.next())
        if(packet->type == OWL::CommPacket::RX)
          {
            OWL::RXPacket *h = (OWL::RXPacket*)packet->data;
            if(*h <= 0) continue;

            int led0 = (((int) h->rxMeta[0]) << 8) | h->rxMeta[1];
            int led_last = (((int) h->rxMeta[2]) << 8) | h->rxMeta[3];
            //printf("0x%x 0x%x 0x%x 0x%x\n", h->rxMeta[0], h->rxMeta[1], h->rxMeta[2], h->rxMeta[3]);
            std::vector<StylusWrapper>::iterator s = find(stylus.begin(), stylus.end(), *h);
            if(s != stylus.end()) continue;

            if(h->type() == OWL::RXPacket::RFScan)
              {
                OWL::RFScanPacket* p = (OWL::RFScanPacket*) h->data;
                if(h->id > 0 && (p->rxAddr[0] | p->rxAddr[1]) && led0 > -1 && led0+1 <= led_last)
                  {
                    printf("new stylus: id=%d hwid=0x%x leds: %d %d\n", (int)h->id, (int) *h, led0, led0+1);
                    stylus.push_back(StylusWrapper(*h, h->id));
                    stylus.back().markers[0] = led0;
                    stylus.back().markers[1] = led0+1;
                  }
              }
          }
    }

  if(stylus.empty()) return 0;

  // incrementally update stylus data
  timeval t0, t1;
  vrpn_gettimeofday(&t0, NULL);
  owlHybridUpdate((const unsigned char*)&data[0], data.size(), &markers[0], markers.size(), &stylus[0], stylus.size());
  vrpn_gettimeofday(&t1, NULL);
  statistics.t2 += vrpn_TimevalDurationSeconds(t1, t0);

  return 1;
}


// ctor
vrpn_Tracker_PhaseSpace::vrpn_Tracker_PhaseSpace(const char *name, vrpn_Connection *c)
  : vrpn_Tracker(name ,c),
    vrpn_Button_Filter(name, c),
    vrpn_Clipping_Analog_Server(name, c),
    numRigids(0),
    numMarkers(0),
    owlRunning(false),
    scale(MM_TO_METERS),
    frame(0),
    frequency(OWL_MAX_FREQUENCY),
    readMostRecent(0),
    slave(0)
{
#ifdef DEBUG
  printf("%s %s\n", __PRETTY_FUNCTION__, name);
#endif

  num_buttons = vrpn_BUTTON_MAX_BUTTONS;
  memset(&statistics, 0, sizeof(statistics));

  if(d_connection) {
    // Register a handler for the update change callback
    if (register_autodeleted_handler(update_rate_id, handle_update_rate_request, this, d_sender_id))
      fprintf(stderr,"vrpn_Tracker: Can't register workspace handler\n");
  }

  memset(&commtime, 0, sizeof(commtime));

  // preallocate buffers for incoming data
  markers.reserve(MAXMARKERS);
  rigids.reserve(MAXRIGIDS);
  commdata.resize(64*1024);
}


// dtor
vrpn_Tracker_PhaseSpace::~vrpn_Tracker_PhaseSpace()
{
#ifdef DEBUG
  printf("%s\n", __PRETTY_FUNCTION__);
#endif
  owlDone();
}


//
bool vrpn_Tracker_PhaseSpace::InitOWL()
{
  printf("connecting to OWL server at %s...\n", device.c_str());

  if(slave) printf("slave mode enabled.\n");
  if(readMostRecent) printf("drop_frames enabled\n");

  // connect to Impulse server
  int flag = slave ? OWL_SLAVE : 0;
  int ret = owlInit(device.c_str(), flag);
  if(ret < 0) {
    fprintf(stderr, "owlInit error: 0x%x\n", ret);
    owlRunning = false;
    return false;
  } else {
    owlRunning = true;
  }

  // query and print server version
  char msg[512];
  if(owlGetString(OWL_VERSION,msg)) {
    printf("OWL version: %s\n",msg);
  } else {
    printf("Unable to query OWL version.\n");
  }

  // enable the data types we want
  owlSetInteger(OWL_EVENTS , OWL_ENABLE);
  owlSetInteger(OWL_MARKERS, OWL_ENABLE);
  owlSetInteger(OWL_RIGIDS , OWL_ENABLE);
  owlSetInteger(OWL_COMMDATA , OWL_ENABLE);

  // set max buffer size to arbitrary maximum
  owlSetInteger(OWL_FRAME_BUFFER_SIZE, 0);

  // set streaming frequency
  setFrequency(frequency);

  // for stylus support
  owlHybridInit(frequency, 1);

  {
    owlGetStatus();
    if((ret = owlGetError()) != OWL_NO_ERROR) {
      fprintf(stderr,"owlhybrid init error: 0x%x\n", ret);
      return false;
    }
  }

  std::vector<int> pt;
  std::vector<int> rt;
  int err;

  if(!slave)
    {
      // create rigid trackers
      for(Sensors::iterator s = sensors.begin(); s != sensors.end(); s++)
        {
          if(s->type == RIGID_BODY)
            {
              numRigids++;
              if(numRigids > MAXRIGIDS)
                {
                  fprintf(stderr, "max rigids exceeded\n");
                  return false;
                }

              if(find(rt.begin(), rt.end(), s->tracker) == rt.end())
                {
                  printf("creating rigid tracker %d\n", s->tracker);
                  owlTrackeri(s->tracker, OWL_CREATE, OWL_RIGID_TRACKER);
                  rt.push_back(s->tracker);

                  // kalman paramters
                  if(s->k[0] || s->k[1] || s->k[2] || s->k[3])
                    {
                      printf("set kalman parameters: %f %f %f %f\n", s->k[0], s->k[1], s->k[2], s->k[3]);
                      owlTrackerfv(s->tracker, OWL_SET_FILTER, s->k);
                    }

                  owlGetStatus();
                  if((err = owlGetError()) != OWL_NO_ERROR) {
                    fprintf(stderr, "libowlsock error 0x%x while adding sensor %d\n", err, s->sensor_id);
                    return false;
                  }
                }
              else
                {
                  fprintf(stderr, "rigid tracker %d already defined\n", s->tracker);
                  return false;
                }
            }
        }

      // leds
      for(Sensors::iterator s = sensors.begin(); s != sensors.end(); s++)
        {
          if(s->type == LED)
            {
              numMarkers++;
              if(numMarkers > MAXMARKERS)
                {
                  fprintf(stderr, "max markers exceeded\n");
                  return false;
                }

              bool rigid = find(rt.begin(), rt.end(), s->tracker) != rt.end();
              if((!rigid) && find(pt.begin(), pt.end(), s->tracker) == pt.end())
                {
                  printf("creating point tracker %d\n", s->tracker);
                  // create point tracker
                  owlTrackeri(s->tracker, OWL_CREATE, OWL_POINT_TRACKER);
                  pt.push_back(s->tracker);
                }

              printf("adding led %d to tracker %d\n", s->id, s->tracker);
              owlMarkeri(MARKER(s->tracker, s->id), OWL_SET_LED, s->id);

              if(rigid)
                {
                  // set rigid body position
                  float p[3];
                  p[0] = s->x;
                  p[1] = s->y;
                  p[2] = s->z;
                  printf("set led position: %f %f %f\n", p[0], p[1], p[2]);
                  owlMarkerfv(MARKER(s->tracker, s->id), OWL_SET_POSITION, p);
                  continue;
                }

              owlGetStatus();
              if((err = owlGetError()) != OWL_NO_ERROR) {
                fprintf(stderr, "libowlsock error 0x%x while adding sensor %d\n", err, s->sensor_id);
                return false;
              }
            }
        }

      // stylus
      for(Sensors::iterator s = sensors.begin(); s != sensors.end(); s++)
        {
          switch(s->type)
            {
            case STYLUS:
              if(s->id < 0 || s->id >= MAX_STYLUS)
                {
                  fprintf(stderr, "error: stylus id must be between 0 and %d\n", MAX_STYLUS);
                  return false;
                }
              break;
            case STYLUS_BUTTON:
            case STYLUS_ANALOG:
            case STYLUS_AIMON:
              if(!check_stylus_defined(s->id))
                {
                  fprintf(stderr, "error while defining sensor %d, stylus %d not defined\n", s->sensor_id, s->id);
                  return false;
                }
              setClipValues(s->sensor_id, s->clip_min, s->clip_low_zero, s->clip_high_zero, s->clip_max);
              break;
            }
        }
    }

  printf("owl init\n");
  return true;
}


//
bool vrpn_Tracker_PhaseSpace::isRunning() const
{
  return owlRunning;
}


// This function should be called each time through the main loop
// of the server code. It polls for data from the OWL server and
// sends them if available.
void vrpn_Tracker_PhaseSpace::mainloop()
{
  timeval t0, t1;
  vrpn_gettimeofday(&t0, NULL);

  statistics.iter1++;

#ifdef BENCHMARK
  timeval t0;
  vrpn_gettimeofday(&t0, NULL);
#endif

  get_report();

#ifdef BENCHMARK
  timeval t1;
  vrpn_gettimeofday(&t1, NULL);
  printf("dt=%d\n", (t1.tv_sec - t0.tv_sec)*1000000 + (((int) t1.tv_usec) - t0.tv_usec));
#endif

  // make sure we're not overflowing
  timeval dt;
  vrpn_gettimeofday(&dt, NULL);
  if(dt.tv_sec != monitortime.tv_sec) //roughly once per second.
    {
      vrpn_gettimeofday(&monitortime, NULL);
      check_stats();
    }

  // Call the generic server mainloop, since we are a server
  server_mainloop();

  vrpn_gettimeofday(&t1, NULL);
  statistics.t0 += vrpn_TimevalDurationSeconds(t1, t0);
  return;
}


// The Impulse server will stop sending data if client side buffers
// fill up.  Monitor and inform the user if data is arriving too fast
void vrpn_Tracker_PhaseSpace::check_stats()
{
  // grab buffer info from server
  char status[1024];
  status[0] = 0;
  owlGetString(OWL_FRAME_BUFFER_SIZE, status);
  //printf("status: %s\n", status);

  // number of checks before a monotonic increase should be reported.
  int maxwait = 4;

  // parse status string for buffer sizes
  int size = 0;
  int events = 0;
  int markers = 0;
  int rigids = 0;
  int strings = 0;
  const char* delim = " ";
  char* token = status;
  char* str = status;
  while(token)
    {
      token = strtok(str, delim);
      if(!token) break;
      str = NULL;
      // if missing from status string,  then buffer is empty.
      if(sscanf(token, "size=%d", &size) > 0) continue;
      if(sscanf(token, "events=%d", &events) > 0) continue;
      if(sscanf(token, "markers=%d", &markers) > 0) continue;
      if(sscanf(token, "rigids=%d", &rigids) > 0) continue;
      if(sscanf(token, "strings=%d", &strings) > 0) continue;
    }

  //printf("%d %d %d %d\n", events, markers, rigids, strings);

#define CHECK(A) if(A > statistics.A.count) statistics.A.delta++;       \
  else statistics.A.delta--;                                            \
  statistics.A.count = A;                                               \
  if(statistics.A.delta < 0) statistics.A.delta = 0;                    \
  if(statistics.A.delta > maxwait || (size > 0 && statistics.A.count >= size)) { \
    printf("warning: " #A " buffer overflow: %d\n", statistics.A.count); \
    statistics.A.delta = 0;                                             \
  }

  // report overflow situations if any
  CHECK(events);
  CHECK(markers);
  CHECK(rigids);
  CHECK(strings);
#undef CHECK

  //printf("events=%d markers=%d rigids=%d strings=%d iter1=%d iter2=%d t0=%2.6f t1=%2.6f t2=%2.6f\n", events, markers, rigids, strings, statistics.iter1, statistics.iter2, statistics.t0, statistics.t1, statistics.t2);

  statistics.iter1 = 0;
  statistics.iter2 = 0;
  statistics.t0 = 0;
  statistics.t1 = 0;
  statistics.t2 = 0;
}


//
std::string trim(char* line, int len)
{
  // cut off leading whitespace
  int s, e;
  for(s = 0; isspace(line[s]) && s < len; s++);
  // cut off comments and trailing whitespace
  e = s;
  for(int i = s; line[i] && i < len; i++)
    {
      if(line[i] == '#') break;
      if(!isspace(line[i])) e = i+1;
    }
  return std::string(line+s, e-s);
}


//
bool read_int(const char* str, int &i)
{
  if(!str) return false;
  errno = 0;
  char* endptr = NULL;
  i = strtol(str, &endptr, 10);
  if(*endptr || errno) {
    fprintf(stderr, "Error, expected an integer but got token: \"%s\"\n", str);
    return false;
  }
  return true;
}


//
bool read_bool(const char* str, bool &b)
{
  if(!str) return false;
  std::string s(str);
  if(s == "true") {
    b = true;
    return true;
  } else if(s == "false") {
    b = false;
    return true;
  }
  int i = 0;
  bool ret = read_int(str, i);
  b = i ? true : false;
  return ret;
}


//
bool read_float(const char* str, float &f)
{
  if(!str) return false;
  errno = 0;
  char* endptr = NULL;
  f = strtod(str, &endptr);
  if(*endptr || errno) {
    fprintf(stderr, "Error, expected a float but got token: \"%s\"\n", str);
    return false;
  }
  return true;
}


//
// TODO: Replace with <regex> once it works in gcc
//
class ConfigParser
{
public:
  std::map <std::string, std::string>  keyvals;

  // returns: 0 on success, 1 on key error
  int pop(std::string key, std::string &val, bool verbose = true)
  {
    if(keyvals.find(key) == keyvals.end()) {
      if(verbose) fprintf(stderr, "Unable to find \"%s\" key.\n", key.c_str());
      return 1;
    }
    val = keyvals[key];
    keyvals.erase(key);
    return 0;
  }

  // returns: 0 on success, 1 on key error, 2 on parse error
  int pop_int(std::string key, int &n, bool verbose = true)
  {
    std::string v;
    if(pop(key, v, verbose)) return 1;
    return read_int(v.c_str(), n) ? 0 : 2;
  }

    // returns: 0 on success, 1 on key error, 2 on parse error
  int pop_float(std::string key, float &n, bool verbose = true)
  {
    std::string v;
    if(pop(key, v, verbose)) return 1;
    return read_float(v.c_str(), n) ? 0 : 2;
  }

    // returns: 0 on success, 1 on key error, 2 on parse error
  int pop_bool(std::string key, bool &n, bool verbose = true)
  {
    std::string v;
    if(pop(key, v, verbose)) return 1;
    return read_bool(v.c_str(), n) ? 0 : 2;
  }

  // return index of error, or -1 on success
  int parse_spec(std::string str)
  {
    std::vector<char> current_key;
    std::vector<char> current_val;

    int i = 0;
    while(str[i])
      {
        current_key.clear();
        current_val.clear();
        while(isspace(str[i])) i++;
        while(str[i]) {
          if(str[i] == '"') {
            fprintf(stderr, "key names are not allowed to contain quotes or be contained in quotes.\n");
            return i;
          } else if(isspace(str[i])) {
            fprintf(stderr, "unexpected whitespace.\n");
            return i;
          } else if(str[i] == '=') {
            i++;
            break;
          } else current_key.push_back(str[i++]);
        }
        if(!current_key.size()) {
          fprintf(stderr, "empty key name.\n");
          return i;
        }
        bool quoted = false;
        if(str[i] == '"') {
          quoted = true;
          i++;
        }
        while(str[i]) {
          if(str[i] == '"') {
            if(quoted) {
              quoted = false;
              i++;
              break;
            } else {
              fprintf(stderr, "misplaced quotes.\n");
              return i;
            }
          } else if(str[i] == '=') {
            fprintf(stderr, "unexpected '=' char in value token.\n");
            return i;
          } else if(!quoted && isspace(str[i])) break;
          else current_val.push_back(str[i++]);
        }
        if(quoted) {
          fprintf(stderr, "unterminated quotes.\n");
          return i;
        }
        if(str[i] && !isspace(str[i])) {
          fprintf(stderr, "expected whitespace after value token.\n");
          return i;
        }
        if(!current_val.size()) {
          fprintf(stderr, "empty value string.\n");
          return i;
        }
        std::string key = std::string(&current_key[0], current_key.size());
        std::string val = std::string(&current_val[0], current_val.size());

        if(keyvals.find(key) != keyvals.end())
          {
            fprintf(stderr, "duplicate key encountered: \"%s\"\n", key.c_str());
            return i;
          }
        keyvals[key] = val;
      }

    return -1;
  }
};


//
bool vrpn_Tracker_PhaseSpace::load(FILE* file)
{
  const int BUFSIZE = 1024;
  char line[BUFSIZE];
  char token[BUFSIZE];
  bool inTag = false;

  std::string ln;
  bool eof = false;

  while(!(eof = (fgets(line, BUFSIZE, file) == NULL)))
    {
      ConfigParser parser;
      ln = trim(line, BUFSIZE);

      // skip contentless lines
      if(!ln.length()) continue;

      // skip lines unless we're nested in <owl> tags
      if(ln == "<owl>") {
        if (inTag) {
          fprintf (stderr, "Error, nested <owl> tag encountered.  Aborting...\n");
          return false;
        } else {
          inTag = true;
          continue;
        }
      } else if (ln == "</owl>") {
        if (inTag) {
          // closing tag encountered,  exit.
          return true;
        } else {
          fprintf (stderr, "Error, </owl> tag without <owl> tag.  Aborting...\n");
          return false;
        }
      }

      // parse line for key-value pairs
      if(inTag) {
        int e = parser.parse_spec(ln);
        if(e >= 0) {
          fprintf(stderr, "Error at character %d.\n", e);
          break;
        }
      }


#define GET(key, func, destination) {                                   \
            int ret = parser.func(key, destination);                    \
            if(ret == 1) {                                              \
              fprintf(stderr, "sensor %d of type \'%s\' missing key \'%s\'\n", info. sensor_id,v.c_str(), key); \
              break;                                                    \
            } else if(ret == 2) {                                       \
              break;                                                    \
            }                                                           \
          }
#define GET_OPT(key, func, destination) if(parser.func(key, destination, false) == 2) break;

      // parse key-value pairs
      std::string k;
      std::string v;
      if(parser.keyvals.find("sensor") != parser.keyvals.end())
        {
          // line is a sensor specification
          SensorInfo info;
          if(parser.pop_int("sensor", info.sensor_id)) break;
          if(parser.pop("type", v)) break;

          if(check_duplicate_sensor(info.sensor_id))
            break;

          if(v == "point") {
            info.type = LED;
            GET("tracker", pop_int, info.tracker);
            GET("led", pop_int, info.id);
            GET_OPT("x", pop_float, info.x);
            GET_OPT("y", pop_float, info.y);
            GET_OPT("z", pop_float, info.z);
          } else if(v == "rigid_body") {
            info.type = RIGID_BODY;
            GET("tracker", pop_int, info.tracker);
            info.id = info.tracker;
            GET_OPT("k0", pop_float, info.k[0]);
            GET_OPT("k1", pop_float, info.k[1]);
            GET_OPT("k2", pop_float, info.k[2]);
            GET_OPT("k3", pop_float, info.k[3]);
            //TODO implement
            //GET_OPT("file", pop, info.file);
          } else if(v == "stylus") {
            info.type = STYLUS;
            GET("stylus_id", pop_int, info.id);

          } else if(v == "stylus_button" ||
                    v == "stylus_analog" ||
                    v == "stylus_aimon") {
            if(v == "stylus_button") info.type = STYLUS_BUTTON;
            if(v == "stylus_analog") info.type = STYLUS_ANALOG;
            if(v == "stylus_aimon") info.type = STYLUS_AIMON;
            GET("stylus_id", pop_int, info.id);
            GET("button_id", pop_int, info.id2);
            GET_OPT("clip_min",      pop_int, info.clip_min);
            GET_OPT("clip_low_zero", pop_int, info.clip_low_zero);
            GET_OPT("clip_high_zero",pop_int, info.clip_high_zero);
            GET_OPT("clip_max",      pop_int, info.clip_max);
          } else {
            fprintf(stderr,"invalid type: %s\n", v.c_str());
            break;
          }

          // check remaining kv's
          if(parser.keyvals.begin() != parser.keyvals.end())
            {
              fprintf(stderr, "invalid key in sensor specification: \"%s\"\n", parser.keyvals.begin()->first.c_str());
              break;
            }

          //
          sensors.push_back(info);
        }
      else
        {
          // line is an option specification
          GET_OPT("device", pop, device);
          GET_OPT("frequency", pop_float, frequency);
          GET_OPT("slave", pop_bool, slave);
          GET_OPT("drop_frames", pop_bool, readMostRecent);
          GET_OPT("scale", pop_float, scale);
          if(parser.keyvals.size() > 0) fprintf(stderr, "unexpected key '%s' in options specification\n", parser.keyvals.begin()->first.c_str());
        }
#undef GET
#undef GET_OPT
    }


  if(eof) fprintf(stderr, "Unexpected end of file.\n");
  else fprintf(stderr, "Unable to parse line: \"%s\"\n", ln.c_str());
  return false;
}



/*
  Enables all the trackers and sets the streaming frequency.
  Note: Trackers according to VRPN and Trackers as defined by OWL
  are two entirely different things.
*/
bool vrpn_Tracker_PhaseSpace::enableStreaming(bool enable)
{
#ifdef DEBUG
  printf("%s %d\n", __PRETTY_FUNCTION__, enable);
#endif

  if(!owlRunning) return false;

  // scale output
  owlScale(scale);

  if(enable) {
    owlSetInteger(OWL_STREAMING, OWL_ENABLE);
    printf("Streaming enabled.\n");
  } else {
    owlSetInteger(OWL_STREAMING, OWL_DISABLE);
    printf("Streaming disabled.\n");
  }

  std::vector<int> trackers;

  // enable/disable the trackers
  if(!slave) {
    for(Sensors::iterator s = sensors.begin(); s != sensors.end(); s++)
      {
        if(find(trackers.begin(), trackers.end(), s->tracker) != trackers.end())
          continue;

        int option = enable ? OWL_ENABLE : OWL_DISABLE;
        if(s->type == RIGID_BODY || s->type == LED)
          {
            printf("tracker %d: %s\n", s->tracker, enable ? "enable" : "false");
            owlTracker(s->tracker, option);
            trackers.push_back(s->tracker);
          }

        {
          owlGetStatus();
          int err = owlGetError();
          if(err != OWL_NO_ERROR) {
            fprintf(stderr, "libowlsock error: 0x%x, line %d\n", err, __LINE__);
            return false;
          }
        }
      }
  }

  {
    owlGetStatus();
    int err = owlGetError();
    if(err != OWL_NO_ERROR) {
      fprintf(stderr, "libowlsock error: 0x%x, line %d\n", err, __LINE__);
      return false;
    }
  }

  return true;
}

//
int vrpn_Tracker_PhaseSpace::read_event(void)
{
  statistics.iter2++;

  int ret = 0;
  OWLEvent e = owlGetEvent();
  switch(e.type) {
  case 0:
    break;
  case OWL_FREQUENCY:
    {
      float freq = 0;
      owlGetFloatv(OWL_FREQUENCY, &freq);
      fprintf(stdout, "frequency: %f\n", freq);
      if(freq != frequency)
        fprintf(stderr, "Error: requested frequency %f Hz, server reported %f Hz\n", frequency, freq);
      frequency = freq;
    }
    break;
  case OWL_MARKERS:
    markers.resize(MAXMARKERS);
    ret = owlGetMarkers(&markers.front(), markers.size());
    if(ret >= 0) markers.resize(ret);
    break;
  case OWL_RIGIDS:
    rigids.resize(MAXRIGIDS);
    ret = owlGetRigids(&rigids.front(), rigids.size());
    if(ret >= 0) rigids.resize(ret);
    break;
  case OWL_COMMDATA:
    // stylus and button data come through COMMDATA interface
    commdata.resize(COMMDATASIZE);

    ret = owlGetString(OWL_COMMDATA, (char*) &commdata[0]);

    if(ret >= 0) {
      timeval t0, t1;
      vrpn_gettimeofday(&t0, NULL);
      commdata.resize(ret);
      if(process_basedata(commdata, frequency, commtime));
      else if(process_hubdata(commdata, frequency, commtime)) {
        update_hybrid(commdata);
      }
      vrpn_gettimeofday(&t1, NULL);
      statistics.t1 += vrpn_TimevalDurationSeconds(t1, t0);

      // remove later
#if 0
      printf("fr=%d commdata=\"");
      for(int i = 0; i < commdata.size(); i++)
        {
          printf("%02x", 0xff & (int) commdata[i]);
        }
      printf("\"\n");
#endif

    }
    break;
  case OWL_TIMESTAMP:
    owlGetIntegerv(OWL_TIMESTAMP, &ret);
    break;
  case OWL_PLANES:
    planes.resize(MAXPLANES);
    ret = owlGetPlanes(&planes.front(), planes.size());
    if(ret >= 0) planes.resize(ret);
    break;
  case OWL_DETECTORS:
    detectors.resize(MAXDETECTORS);
    ret = owlGetDetectors(&detectors.front(), detectors.size());
    if(ret >= 0) detectors.resize(ret);
    break;
  case OWL_PEAKS:
    peaks.resize(MAXPEAKS);
    ret = owlGetPeaks(&peaks.front(), peaks.size());
    if(ret >= 0) peaks.resize(ret);
    break;
  case OWL_IMAGES:
    images.resize(MAXIMAGES);
    ret = owlGetImages(&images.front(), images.size());
    if(ret >= 0) images.resize(ret);
    break;
  case OWL_CAMERAS:
    cameras.resize(MAXCAMERAS);
    ret = owlGetCameras(&cameras.front(), cameras.size());
    if(ret >= 0) cameras.resize(ret);
    break;
  case OWL_STATUS_STRING: break;
  case OWL_CUSTOM_STRING: break;
  case OWL_FRAME_NUMBER:
    if(owlGetIntegerv(OWL_FRAME_NUMBER, &ret))
      frame = ret;
    break;
  case OWL_DONE:
    owlRunning = false;
    // inform client we are done
    send_text_message("owl stopped\n", vrpn_Tracker::timestamp, vrpn_TEXT_ERROR);
    owlDone();
    return 0;
  default:
    fprintf(stderr, "unsupported OWLEvent type: 0x%x\n", e.type);
    break;
  }
  return e.type;
}

//
void vrpn_Tracker_PhaseSpace::set_pose(const OWLRigid &r)
{
  //set the position
  pos[0] = r.pose[0];
  pos[1] = r.pose[1];
  pos[2] = r.pose[2];

  //set the orientation quaternion
  //OWL has the scale factor first, whereas VRPN has it last.
  d_quat[0] = r.pose[4];
  d_quat[1] = r.pose[5];
  d_quat[2] = r.pose[6];
  d_quat[3] = r.pose[3];
}

//
void vrpn_Tracker_PhaseSpace::report_marker(vrpn_int32 sensor, const OWLMarker& m)
{
  if(m.cond <= 0) return;
  d_sensor = sensor;

  pos[0] = m.x;
  pos[1] = m.y;
  pos[2] = m.z;

  //raw positions have no rotation
  d_quat[0] = 0;
  d_quat[1] = 0;
  d_quat[2] = 0;
  d_quat[3] = 1;

#ifdef DEBUG
  printf("%d marker fr=%d id=%d:%d cond=%f x=%f y=%f z=%f\n", d_sensor, m.frame, TRACKER(m.id), INDEX(m.id), m.cond, m.x, m.y, m.z);
#endif

  send_report();
}


//
void vrpn_Tracker_PhaseSpace::report_rigid(vrpn_int32 sensor, const OWLRigid& r, bool is_stylus)
{
  if(r.cond <= 0) return;
  d_sensor = sensor;

  // set the position/orientation
  set_pose(r);

#ifdef DEBUG
  if(is_stylus)
    {
      printf("%d stylus fr=%d id=%d:%d cond=%f x=%f y=%f z=%f w=%f a=%f b=%f c=%f\n", d_sensor, r.frame, TRACKER(r.id), INDEX(r.id), r.cond, r.pose[0], r.pose[1], r.pose[2], r.pose[3], r.pose[4], r.pose[5], r.pose[6]);
    }
  else
    {
      printf("%d rigid fr=%d id=%d:%d cond=%f x=%f y=%f z=%f w=%f a=%f b=%f c=%f\n", d_sensor, r.frame, TRACKER(r.id), INDEX(r.id), r.cond, r.pose[0], r.pose[1], r.pose[2], r.pose[3], r.pose[4], r.pose[5], r.pose[6]);
    }
#endif

  send_report();
}


//
void vrpn_Tracker_PhaseSpace::send_report(void)
{
  if(d_connection)
    {
      char	msgbuf[MSGBUFSIZE];
      int	len = vrpn_Tracker::encode_to(msgbuf);
      if(d_connection->pack_message(len, vrpn_Tracker::timestamp, position_m_id, d_sender_id, msgbuf,
                                    vrpn_CONNECTION_LOW_LATENCY)) {
        fprintf(stderr,"PhaseSpace: cannot write message: tossing\n");
      }
    }
}


//
void vrpn_Tracker_PhaseSpace::report_button(vrpn_int32 sensor, int value)
{
  if(d_sensor < 0 || d_sensor >= vrpn_BUTTON_MAX_BUTTONS)
    {
      fprintf(stderr, "error: sensor %d exceeds max button count\n", sensor);
      return;
    }

  d_sensor = sensor;
  buttons[d_sensor] = value;

#ifdef DEBUG
  printf("send button %d 0x%x\n", d_sensor, value);
#endif
}


//
void vrpn_Tracker_PhaseSpace::report_button_analog(vrpn_int32 sensor, int value)
{
  d_sensor = sensor;
  setChannelValue(d_sensor, value);

#ifdef DEBUG
  printf("send analog button %d 0x%x\n", d_sensor, value);
#endif
}



//
template<class A>
A* find(int id, int& hint, std::vector<A> &data)
{
  if(hint < 0 || hint >= data.size() || id != data[hint].id)
    {
      for(int i = 0; i < data.size(); i++)
        {
          if(data[i].id == id)
            {
              hint = i;
              return &data[i];
            }
        }
      return NULL;
    }
  return &data[hint];
}


//
int vrpn_Tracker_PhaseSpace::get_report(void)
{
  if(!owlRunning) return 0;

#ifdef DEBUG
  char buffer[1024];
  owlGetString(OWL_FRAME_BUFFER_SIZE, buffer);
  printf("%s\n", buffer);
#endif

  int maxiter = 1;

  // set limit on looping
  if(readMostRecent) maxiter = 10000;

  // read new data
  int oldframe = frame;
  int ret = 1;
  for(int i = 0; i < maxiter && ret; i++)
    ret = read_event();

  // no new data? abort.
  if(oldframe == frame)
    return 0;

  frame_count = frame;

  // set timestamp for subsequent reports
  if(frequency)
    {
      if(commtime.tv_sec && commtime.tv_usec) vrpn_Tracker::timestamp = commtime; // use commdata time if present
      else frame_to_timeval(frame, frequency, vrpn_Tracker::timestamp); // calculate timestamp from frame number
    }
  else
    {
      memset(&(vrpn_Tracker::timestamp), 0, sizeof(vrpn_Tracker::timestamp));
    }


  // get stylus scalars
  const int bps = SCALARS_PER_STYLUS;
  std::vector<char> incoming_scalars;
  if(d_connection && stylus.size())
    {
      incoming_scalars.resize(MAXSCALARS);
      owlHybridGetScalars(&stylus[0], stylus.size(), &incoming_scalars[0], incoming_scalars.size(), bps);
#if 0
      printf("buttons ");
      for(int i = 0; i < 16 && i < incoming_scalars.size(); i++) {
        printf("%02x ", 0xff & (unsigned int) incoming_scalars[i]);
      }
      printf("\n");
#endif
    }

  // convert stylus to rigids
  stylus_to_rigids(stylus, stylus_rigids);

  // send reports
  if(slave) {

    // in slave mode,  sensor is determined simply by order of arrival
    int slave_sensor_counter = 0;
    for(int i = 0; i < markers.size(); i++)
      report_marker(slave_sensor_counter++, markers[i]);
    for(int i = 0; i < rigids.size(); i++)
      report_rigid(slave_sensor_counter++, rigids[i]);
    for(int i = 0; i < stylus_rigids.size(); i++)
      {
        report_rigid(slave_sensor_counter++, stylus_rigids[i], true);

        // currently no way to know what devices are attached to stylus,  just spit out the scalars.
        int index = 0;
        if(find(stylus_rigids[i].id, index, stylus))
          for(int j = 0; j < bps; j++)
            {
              setClipValues(slave_sensor_counter, 50, 156, 163, 255);
              report_button_analog(slave_sensor_counter++, 0xff & (int) incoming_scalars[index * bps + j]);
              report_button(slave_sensor_counter++, incoming_scalars[index * bps + j]);
            }
      }

  } else {

    for(Sensors::iterator i = sensors.begin(); i != sensors.end(); i++) {
      switch(i->type)
        {
        case LED:
          {
            OWLMarker* m = find(MARKER(i->tracker, i->id), i->index, markers);
            if(m) report_marker(i->sensor_id, *m);
            break;
          }
        case RIGID_BODY:
          {
            OWLRigid* r = find(i->id, i->index, rigids);
            if(r) report_rigid(i->sensor_id, *r);
            break;
          }
        case STYLUS:
          {
            OWLRigid* r = find(i->id, i->index, stylus_rigids);
            if(r) report_rigid(i->sensor_id, *r, true);
            break;
          }
        case STYLUS_BUTTON:
        case STYLUS_ANALOG:
          {
            StylusWrapper* s = find(i->id, i->index, stylus);
            if(!s) break;
            int index = i->index * bps + i->id2;
            if(index < 0 || index >= incoming_scalars.size())
              break;

            int value = 0xff & (int) incoming_scalars[index];
            if(i->type == STYLUS_BUTTON)
              {
                report_button(i->sensor_id, value ? 1 : 0);
              }
            else
              {
                report_button_analog(i->sensor_id, value);
              }
            break;
          }
        case STYLUS_AIMON:
          {
            StylusWrapper* s = find(i->id, i->index, stylus);
            if(!s) break;
            int index = i->index * bps;
            if(index < 0 || (index+bps) >= incoming_scalars.size())
              break;

            char b;
            switch(i->id2)
              {
              case 0: // joystick x-axis (analog)
                b = incoming_scalars[index+0];
                break;
              case 1: // joystick y-axis (analog)
                b = incoming_scalars[index+1];
                break;
              case 2: // button 1 (analog)
                b = (0x0f & incoming_scalars[index + 2]) * 16;
                break;
              case 3: // button 2 (analog)
                b = (0x0f & ((0xf0 & incoming_scalars[index + 2])>>4)) * 16;
                break;
              case 4: // right side button (boolean)
                b = (0x40 & incoming_scalars[index + 3]) ? 255 : 0;
                break;
              case 5: // left side button (boolean)
                b = (0x20 & incoming_scalars[index + 3]) ? 255 : 0;
                break;
              case 6: // joystick z-axis (boolean)
                b = (0x10 & incoming_scalars[index + 3]) ? 255 : 0;
                break;
              case 7: // dpad right (boolean)
                b = (0x08 & incoming_scalars[index + 3]) ? 255 : 0;
                break;
              case 8: // dpad left (boolean)
                b = (0x04 & incoming_scalars[index + 3]) ? 255 : 0;
                break;
              case 9: // dpad down (boolean)
                b = (0x02 & incoming_scalars[index + 3]) ? 255 : 0;
                break;
              case 10: // dpad up (boolean)
                b = (0x01 & incoming_scalars[index + 3]) ? 255 : 0;
                break;
              }
            report_button_analog(i->sensor_id, 0xff & (int) b);
            report_button(i->sensor_id, b ? 1 : 0);
          }
          break;
        default:
          fprintf(stderr, "error: sensor %d has unknown sensor type: 0x%x\n", i->sensor_id, i->type);
          continue;
        }
    }

  }

  // finalize button and analog reports
  vrpn_Button_Filter::report_changes();
  vrpn_Clipping_Analog_Server::report_changes();

  return markers.size() || rigids.size() || stylus.size() ? 1 : 0;
}



//
void vrpn_Tracker_PhaseSpace::setFrequency(float freq)
{
#ifdef DEBUG
  printf("%s %f\n", __PRETTY_FUNCTION__, freq);
#endif
  if(owlRunning) {
    owlSetFloat(OWL_FREQUENCY, freq);
    owlGetStatus();
    int err = owlGetError();
    if(err != OWL_NO_ERROR)
      {
        fprintf(stderr, "Error: unable to set frequency to %f Hz\n", freq);
        fprintf(stderr, "libowlsock error: 0x%x, line %d\n", err, __LINE__);
      }
    else frequency = freq;
  }
  return;
}

//
bool vrpn_Tracker_PhaseSpace::check_duplicate_sensor(int sensor)
{
  for(Sensors::iterator i = sensors.begin(); i != sensors.end(); i++)
    {
      if(i->sensor_id == sensor) {
        fprintf(stderr, "sensor %d previously defined.\n", sensor);
        return true;
      }
    }
  return false;
}

//
bool vrpn_Tracker_PhaseSpace::check_stylus_defined(int stylus_id) {
  for(Sensors::iterator i = sensors.begin(); i != sensors.end(); i++) {
    if(i->type == STYLUS && i->id == stylus_id)
      return true;
  }
  return false;
}

//
int vrpn_Tracker_PhaseSpace::handle_update_rate_request(void *userdata, vrpn_HANDLERPARAM p)
{
#ifdef DEBUG
  printf("%s\n", __PRETTY_FUNCTION__);
#endif
  vrpn_Tracker_PhaseSpace* thistracker = (vrpn_Tracker_PhaseSpace*)userdata;
  vrpn_float64 update_rate = 0;
  vrpn_unbuffer(&p.buffer,&update_rate);
  thistracker->setFrequency((float) update_rate);
  return 0;
}

#endif
