#ifndef VRPN_TRACKER_PHASESPACE_H
#define VRPN_TRACKER_PHASESPACE_H

#include "vrpn_Configure.h"   // IWYU pragma: keep

#ifdef  VRPN_INCLUDE_PHASESPACE
#include <stdint.h>
#include <map>
#include <vector>
#include <string>

#include "vrpn_Shared.h"
#include "vrpn_Tracker.h"
#include "vrpn_Button.h"
#include "vrpn_Analog.h"

#include "owl.h"
#include "owl_planes.h"
#include "owl_peaks.h"
#include "owl_images.h"
#include "owl_comm.h"
#include "owlhybrid.h"


class VRPN_API vrpn_Tracker_PhaseSpace : public vrpn_Tracker, public vrpn_Button_Filter, public vrpn_Clipping_Analog_Server {

public:

  vrpn_Tracker_PhaseSpace(const char *name,
                          vrpn_Connection *c);

  vrpn_Tracker_PhaseSpace(const char *name,
                          vrpn_Connection *c,
                          const char* device,
                          float frequency,
                          int readflag,
                          int slaveflag=0);


  ~vrpn_Tracker_PhaseSpace();

  // vrpn_Tracker
  virtual void mainloop();
  static int VRPN_CALLBACK handle_update_rate_request(void *userdata, vrpn_HANDLERPARAM p);

  // check if connected to PhaseSpace server
  bool isRunning() const;

  // parse a tracker specification file and create PhaseSpace trackers
  bool load(FILE* file);

  // connect to the Impulse system
  bool InitOWL();

  // set fps
  void setFrequency(float freq);

  // start streaming
  bool enableStreaming(bool enable);

  // sensor types
  enum { LED, RIGID_BODY, STYLUS, STYLUS_BUTTON, STYLUS_ANALOG, STYLUS_AIMON };

  // wrapper to provide constructor and operators for OWLStylus
  // Must be same size as OWLStylus, do not add members.
  struct StylusWrapper : public OWLStylus {
    StylusWrapper(const StylusWrapper &o);
    StylusWrapper(int hwid, int id, float width);
    bool operator==(int n) const;
  };

protected:

  std::string device;
  bool owlRunning;
  bool readMostRecent;
  bool slave;

  float scale;
  float frequency;
  int numRigids;
  int numMarkers;
  int64_t frame;
  struct timeval monitortime;
  struct timeval commtime;

  // for monitoring overflow
  struct {
    struct {
      int count;
      int delta;
    } events;
    struct {
      int count;
      int delta;
    } markers;
    struct {
      int count;
      int delta;
    } rigids;
    struct {
      int count;
      int delta;
    } strings;

    int iter1;
    int iter2;

    double t0;
    double t1;
    double t2;

  } statistics;

  //
  struct SensorInfo {
    int index; // memoized index for faster find
    vrpn_int32 sensor_id; // vrpn sensor
    int type;

    int tracker;
    int id;
    int id2;

    float x, y, z;
    float k[4];

    int clip_min, clip_low_zero, clip_high_zero, clip_max;

    SensorInfo()
    {
      memset(this, 0, sizeof(SensorInfo));
      clip_min = 50;
      clip_low_zero = 156;
      clip_high_zero = 163;
      clip_max = 255;      
    }
  };

  // mapping info for vrpn sensors to PhaseSpace ids
  typedef std::vector<SensorInfo> Sensors;
  Sensors sensors;

  // local storage for PhaseSpace data
  std::vector<OWLMarker> markers;
  std::vector<OWLRigid> rigids;
  std::vector<OWLCamera> cameras;
  std::vector<OWLPlane> planes;
  std::vector<OWLPeak> peaks;
  std::vector<OWLImage> images;
  std::vector<OWLDetectors> detectors;
  std::vector<unsigned char> commdata;

  std::vector<StylusWrapper> stylus;
  std::vector<OWLRigid> stylus_rigids;

protected:

  // functions to create and edit PhaseSpace trackers
  // bool addMarker(int sensor,int led_id);
  // bool startNewRigidBody(int sensor, float* kalman=NULL, int kcount=0);
  // bool addRigidMarker(int sensor, int led_id, float x, float y, float z);
  // bool addStylus(int sensor, int id);
  // bool addStylusButton(int sensor, int stylus_id, int id);
  // bool addStylusAnalog(int sensor, int stylus_id, int id);
  // bool addStylusButtonAimon(int sensor, int stylus_id, int id);

  bool check_duplicate_sensor(int sensor);;
  bool check_stylus_defined(int stylus_id);
  void check_stats();

  int read_event(void);

  int update_hybrid(const std::vector<unsigned char> &commdata);

  void set_pose(const OWLRigid &r);

  void report_marker(vrpn_int32 sensor, const OWLMarker &m);
  void report_rigid(vrpn_int32 sensor, const OWLRigid &r, bool is_stylus=false);
  void report_button(vrpn_int32 sensor, int value);
  void report_button_analog(vrpn_int32 sensor, int value);



  // vrpn_Tracker
  virtual int get_report(void);
  virtual void send_report(void);
};
#endif

#endif
